@@ -50,7 +50,8 @@ let evalIdGen = 0;
 
 
 class FuzzySearchResults {
-    constructor(element, entries) {
+    constructor(repl, element, entries) {
+        this.repl = repl;
         this.element = element;
         this.entries = entries;
         this.structured_results = null;
@@ -77,7 +78,13 @@ class FuzzySearchResults {
         this.results.innerHTML = "";
         for (const result of results) {
             const row = document.createElement("div");
-            row.innerText = this.entries[result.item];
+            row.classList.add("snek-fuzzy-entry");
+            const txt = this.entries[result.item];
+            row.innerText = txt;
+            row.onclick = () => {
+                this.repl.selectHistoryEntry(txt);
+                this.repl.editor.focus();
+            };
             this.results.appendChild(row);
         }
         if (!this.results.children.length) {
@@ -134,6 +141,7 @@ class Repl {
     constructor(target) {
         this.container = target;
         this.options = {};
+        this.lib = {};
         this.pane = target.querySelector(".snek-pane");
         this.outerPane = target.querySelector(".snek-outer-pane");
         this.pinpane = target.querySelector(".snek-pin-pane");
@@ -141,6 +149,7 @@ class Repl {
         this.inputBox = target.querySelector(".snek-input");
         this.inputMode = target.querySelector(".snek-input-mode");
         this.statusBar = target.querySelector(".snek-status-bar");
+        this.statusBar.onclick = () => this.cmd_status("/status")
         this.nav = target.querySelector(".snek-nav");
         this.statusHistory = document.createElement("div");
         this.statusHistory.className = "snek-status-history";
@@ -156,6 +165,7 @@ class Repl {
         target.onclick = this.$globalClickEvent.bind(this);
         window.onkeydown = this.$globalKDEvent.bind(this);
         window.$$SKTK = this.sktk.bind(this);
+        window.snektalk = this;
         this.$currid = 0;
         this.$responseMap = {};
         exports.mainRepl = this;
@@ -232,9 +242,8 @@ class Repl {
         }
     }
 
-    sktk(id, ...args) {
-        // Call a Python function by id
-        const exec = async (...args) => {
+    get_external(id) {
+        return async (...args) => {
             let response_id = this.$currid++;
             let response = new Promise(
                 (resolve, reject) => {
@@ -249,6 +258,11 @@ class Repl {
             });
             return await response;
         }
+    }
+
+    sktk(id, ...args) {
+        // Call a Python function by id
+        const exec = this.get_external(id);
 
         const execNow = async () => {
             try {
@@ -419,15 +433,7 @@ class Repl {
 
         editor.addCommand(
             KC.Enter,
-            () => {
-                const selection = this.historyPopup.get();
-                if (selection) {
-                    this.editor.setValue(selection);
-                    let nlines = this.editor.getModel().getLineCount();
-                    this.editor.setPosition({lineNumber: nlines, column: 1000000});
-                }
-                this.destroyHistoryPopup();
-            },
+            () => this.selectHistoryEntry(),
             "popupActive"
         );
 
@@ -481,12 +487,43 @@ class Repl {
             () => { this.editor.trigger(null, "editor.action.commentLine"); },
         );
 
+        editor.addCommand(
+            KM.chord(KM.CtrlCmd | KC.KEY_K, KM.CtrlCmd | KC.KEY_K),
+            () => {
+                this.editor.setValue("");
+                this.lib.stop();
+            }
+        );
+
+        if (isMac) {
+            editor.addCommand(
+                KM.WinCtrl | KC.KEY_C,
+                () => {
+                    this.editor.setValue("");
+                    this.lib.stop();
+                }
+            );
+        }
+
         this.editor = editor;
     }
 
+    selectHistoryEntry(sel) {
+        const selection = sel || this.historyPopup.get();
+        console.log(sel);
+        console.log(selection);
+        if (selection) {
+            this.editor.setValue(selection);
+            let nlines = this.editor.getModel().getLineCount();
+            this.editor.setPosition({lineNumber: nlines, column: 1000000});
+        }
+        this.destroyHistoryPopup();
+    }
+
     setupHistoryPopup() {
         const el = document.createElement("div");
         const fuzz = new FuzzySearchResults(
+            this,
             el,
             this.history.slice(1),
         );
@@ -678,6 +715,33 @@ class Repl {
         }
     }
 
+    recv_fill(data) {
+        let elem = this.reify(data.value);
+        let target = document.getElementById(data.target);
+        while (target.firstChild) {
+            target.removeChild(target.firstChild);
+        }
+        target.appendChild(elem);
+    }
+
+    recv_insert(data) {
+        let elem = this.reify(data.value);
+        let target = document.getElementById(data.target);
+        if (data.index === undefined || data.index === null) {
+            target.appendChild(elem);
+        }
+        else {
+            target.insertBefore(elem, target.children[data.index]);
+        }
+    }
+
+    recv_clear(data) {
+        let target = document.getElementById(data.target);
+        while (target.firstChild) {
+            target.removeChild(target.firstChild);
+        }
+    }
+
     recv_echo(data) {
         let ed = this.read_only_editor(data.value.trimEnd(), data.language || "python");
         let elem = document.createElement("div");
@@ -721,6 +785,28 @@ class Repl {
         this.inputMode.innerHTML = data.html;
     }
 
+    recv_add_history(data) {
+        this.history.push.apply(this.history, data["history"].reverse());
+    }
+
+    recv_set_lib(data) {
+        for (let key in data.lib) {
+            this.lib[key] = this.get_external(data.lib[key]);
+        }
+    }
+
+    recv_broadcast(data) {
+        let selector = `*[channel-${data.key}=""]`
+        if (data.subkey) {
+            selector = `${selector}, *[channel-${data.key}="${data.subkey}"]`
+        }
+        for (let element of document.querySelectorAll(selector)) {
+            if (element.onbroadcast) {
+                element.onbroadcast(data);
+            }
+        }
+    }
+
     recv_bad(data) {
         console.error("Received an unknown command:", data.command);
     }
