@@ -5,6 +5,7 @@ import inspect
 import json
 import os
 import random
+import re
 import threading
 import traceback
 from collections import defaultdict, deque
@@ -14,24 +15,54 @@ from itertools import count
 
 from hrepr import H, Tag, hrepr
 
+from .config import mayread, maywrite
 from .registry import callback_registry
 
 _c = count(1)
 
 
+class NoPatternException(Exception):
+    pass
+
+
+class CommandDispatcher:
+    def __init__(self, patterns):
+        self.patterns = [
+            (re.compile(pattern, re.MULTILINE | re.DOTALL), fn)
+            for pattern, fn in patterns.items()
+        ]
+
+    def __call__(self, expr):
+        for (pattern, fn) in self.patterns:
+            if m := pattern.fullmatch(expr):
+                try:
+                    return fn(m[0], *m.groups())
+                except NotImplementedError:
+                    continue
+        else:
+            raise NoPatternException(f"'{expr}' matches no known pattern")
+
+
 class ThreadKilledException(Exception):
     pass
 
 
+class SnektalkInterrupt(Exception):
+    pass
+
+
+def kill_thread(thread, exctype=ThreadKilledException):
+    ctypes.pythonapi.PyThreadState_SetAsyncExc(
+        ctypes.c_long(thread.ident), ctypes.py_object(exctype)
+    )
+
+
 class KillableThread(threading.Thread):
     @property
     def dead(self):
         return not self.is_alive() or getattr(self, "_dead", False)
 
-    def kill(self):
-        ctypes.pythonapi.PyThreadState_SetAsyncExc(
-            ctypes.c_long(self.ident), ctypes.py_object(ThreadKilledException)
-        )
+    kill = kill_thread
 
 
 class NamedThreads:
@@ -115,8 +146,32 @@ def new_evalid():
         _current_evalid.reset(token)
 
 
+class Lib:
+    def __init__(self, session):
+        self._export = None
+        self.session = session
+
+    def method_stop(self):
+        thread = self.session.owner or threading.main_thread()
+        kill_thread(thread, SnektalkInterrupt)
+
+    def export(self):
+        if self._export:
+            return self._export
+
+        self._export = rval = {}
+        for method_name in dir(type(self)):
+            if method_name.startswith("method_"):
+                method = getattr(self, method_name)
+                method_id = callback_registry.register(method)
+                rval[method_name[7:]] = method_id
+
+        return rval
+
+
 class Session:
-    def __init__(self, socket=None):
+    def __init__(self, socket=None, history_file=None):
+        self.lib = Lib(self)
         self.blt = vars(builtins)
         self.idmap = {}
         self.varcount = count(1)
@@ -129,6 +184,15 @@ class Session:
         self.semaphores = defaultdict(lambda: threading.Semaphore(value=0))
         self.navs = {}
         self.owners = []
+        self.history_file = history_file
+        self.history = mayread(history_file, default=[])
+        self.dispatch = CommandDispatcher(
+            {
+                "/attach[ \n]?(.*)": self.submit_command_attach,
+                "/detach[ \n]?(.*)": self.submit_command_detach,
+                "/kill[ \n]?(.*)": self.submit_command_kill,
+            }
+        )
         self._token = None
         self._tokenp = None
 
@@ -272,6 +336,8 @@ class Session:
         self.sent_resources = set()
         while self.out_queue:
             self.schedule(self.send(**self.out_queue.popleft()))
+        self.queue(command="add_history", history=self.history)
+        self.queue(command="set_lib", lib=self.lib.export())
         self.submit({"command": "noop"})
 
     async def send(self, process=True, **command):
@@ -324,6 +390,48 @@ class Session:
     # Commands #
     ############
 
+    def submit_command_attach(self, expr, tname):
+        tname = tname.strip()
+        self.queue(command="echo", value=expr, process=False)
+        if tname == "main":
+            self.push_owner(None)
+        elif tname in threads.threads:
+            thread = threads.threads[tname]
+            self.push_owner(thread)
+        else:
+            self.queue(
+                command="result",
+                value=f"No thread named {tname}"
+                if tname
+                else "Please provide the name of the thread to attach to",
+                type="exception",
+            )
+
+    def submit_command_detach(self, expr, arg):
+        self.queue(command="echo", value=expr, process=False)
+        assert not arg.strip()
+        self.pop_owner()
+
+    def submit_command_kill(self, expr, tname):
+        tname = tname.strip()
+        self.queue(command="echo", value=expr, process=False)
+        if tname in threads.threads:
+            thread = threads.threads[tname]
+            thread.kill()
+            self.queue(
+                command="result",
+                value=f"Sent an exception to {tname}. It should terminate as soon as possible.",
+                type="print",
+            )
+        else:
+            self.queue(
+                command="result",
+                value=f"No thread named {tname}"
+                if tname
+                else "Please provide the name of the thread to kill",
+                type="exception",
+            )
+
     def submit(self, data):
         self.in_queue.append(data)
         self.semaphores[self.owner].release()
@@ -334,7 +442,12 @@ class Session:
         await meth(**command)
 
     async def command_submit(self, *, expr):
-        self.submit({"command": "expr", "expr": expr})
+        if expr.strip():
+            self.history.append(expr)
+        try:
+            self.dispatch(expr)
+        except NoPatternException:
+            self.submit({"command": "expr", "expr": expr})
 
     async def command_callback(self, *, id, response_id, arguments):
         try:
@@ -349,7 +462,7 @@ class Session:
 
         try:
             with self.set_context():
-                if inspect.isawaitable(cb):
+                if inspect.isawaitable(cb) or inspect.iscoroutinefunction(cb):
                     result = await cb(*arguments)
                 else:
                     result = cb(*arguments)
@@ -370,3 +483,6 @@ class Session:
                 },
                 response_id=response_id,
             )
+
+    def atexit(self):
+        maywrite(self.history_file, self.history[:1000])
